namespace {

// Groups:
char const * const s_grpDefaultJobSettings                                      = "defaultJobSettings";
char const * const s_grpDefaults                                                = "defaults";
char const * const s_grpDerivingTrackLanguagesFromFileNames                     = "derivingTrackLanguagesFromFileNames";
char const * const s_grpFileColors                                              = "fileColors";
char const * const s_grpHeaderViewManager                                       = "headerViewManager";
char const * const s_grpInfo                                                    = "info";
char const * const s_grpLanguageShortcuts                                       = "languageShortcuts";
char const * const s_grpRunProgramConfigurations                                = "runProgramConfigurations";
char const * const s_grpSettings                                                = "settings";
char const * const s_grpSplitterSizes                                           = "splitterSizes";
char const * const s_grpUpdates                                                 = "updates";
char const * const s_grpWindowGeometry                                          = "windowGeometry";

// Values:
char const * const s_valActive                                                  = "active";
char const * const s_valAudioFile                                               = "audioFile";
char const * const s_valAudioPolicy                                             = "audioPolicy";
char const * const s_valAutoClearFileTitle                                      = "autoClearFileTitle";
char const * const s_valAutoClearOutputFileName                                 = "autoClearOutputFileName";
char const * const s_valAutoDestinationOnlyForVideoFiles                        = "autoDestinationOnlyForVideoFiles";
char const * const s_valAutoSetFileTitle                                        = "autoSetFileTitle";
char const * const s_valBCP47LanguageEditingMode                                = "bcp47LanguageEditingMode";
char const * const s_valBCP47NormalizationMode                                  = "bcp47NormalizationMode";
char const * const s_valBoundaryChars                                           = "boundaryChars";
char const * const s_valCeTextFileCharacterSet                                  = "ceTextFileCharacterSet";
char const * const s_valChapterNameTemplate                                     = "chapterNameTemplate";
char const * const s_valCheckForUpdates                                         = "checkForUpdates";
char const * const s_valChecksums                                               = "checksums";
char const * const s_valClearMergeSettings                                      = "clearMergeSettings";
char const * const s_valColumnOrder                                             = "columnOrder";
char const * const s_valColumnSizes                                             = "columnSizes";
char const * const s_valCommandLine                                             = "commandLine";
char const * const s_valDefaultAdditionalMergeOptions                           = "defaultAdditionalMergeOptions";
char const * const s_valDefaultAudioTrackLanguage                               = "defaultAudioTrackLanguage";
char const * const s_valDefaultChapterLanguage                                  = "defaultChapterLanguage";
char const * const s_valDefaultDeriveCommentaryFlagFromFileNames                = "defaultDeriveCommentaryFlagFromFileNames";
char const * const s_valDefaultDeriveFlagsFromTrackNames                        = "defaultDeriveFlagsFromTrackNames";
char const * const s_valDefaultDeriveHearingImpairedFlagFromFileNames           = "defaultDeriveHearingImpairedFlagFromFileNames";
char const * const s_valDefaultDeriveSubtitlesForcedFlagFromFileNames           = "defaultDeriveSubtitlesForcedFlagFromFileNames";
char const * const s_valDefaultRegexForDerivingCommentaryFlagFromFileNames      = "defaultRegexForDerivingCommentaryFlagFromFileNames";
char const * const s_valDefaultRegexForDerivingHearingImpairedFlagFromFileNames = "defaultRegexForDerivingHearingImpairedFlagFromFileNames";
char const * const s_valDefaultRegexForDerivingSubtitlesForcedFlagFromFileNames = "defaultRegexForDerivingSubtitlesForcedFlagFromFileNames";
char const * const s_valDefaultSetOriginalLanguageFlagLanguage                  = "defaultSetOriginalLanguageFlagLanguage";
char const * const s_valDefaultSubtitleCharset                                  = "defaultSubtitleCharset";
char const * const s_valDefaultSubtitleTrackLanguage                            = "defaultSubtitleTrackLanguage";
char const * const s_valDefaultTrackLanguage                                    = "defaultTrackLanguage";
char const * const s_valDefaultVideoTrackLanguage                               = "defaultVideoTrackLanguage";
char const * const s_valDisableCompressionForAllTrackTypes                      = "disableCompressionForAllTrackTypes";
char const * const s_valDisableDefaultTrackForSubtitles                         = "disableDefaultTrackForSubtitles";
char const * const s_valDropLastChapterFromBlurayPlaylist                       = "dropLastChapterFromBlurayPlaylist";
char const * const s_valElideTabHeaderLabels                                    = "elideTabHeaderLabels";
char const * const s_valEnableMuxingAllAudioTracks                              = "enableMuxingAllAudioTracks";
char const * const s_valEnableMuxingAllSubtitleTracks                           = "enableMuxingAllSubtitleTracks";
char const * const s_valEnableMuxingAllVideoTracks                              = "enableMuxingAllVideoTracks";
char const * const s_valEnableMuxingForcedSubtitleTracks                        = "enableMuxingForcedSubtitleTracks";
char const * const s_valEnableMuxingTracksByLanguage                            = "enableMuxingTracksByLanguage";
char const * const s_valEnableMuxingTracksByTheseLanguages                      = "enableMuxingTracksByTheseLanguages";
char const * const s_valEnableMuxingTracksByTheseTypes                          = "enableMuxingTracksByTheseTypes";
char const * const s_valFixedOutputDir                                          = "fixedOutputDir";
char const * const s_valForEvents                                               = "forEvents";
char const * const s_valGuiVersion                                              = "guiVersion";
char const * const s_valHeaderEditorDateTimeInUTC                               = "headerEditorDateTimeInUTC";
char const * const s_valHeaderEditorDroppedFilesPolicy                          = "headerEditorDroppedFilesPolicy";
char const * const s_valHexDumps                                                = "hexDumps";
char const * const s_valHexPositions                                            = "hexPositions";
char const * const s_valHiddenColumns                                           = "hiddenColumns";
char const * const s_valIgnorePlaylistsForMenus                                 = "ignorePlaylistsForMenus";
char const * const s_valJobRemovalOnExitPolicy                                  = "jobRemovalOnExitPolicy";
char const * const s_valJobRemovalPolicy                                        = "jobRemovalPolicy";
char const * const s_valLanguage                                                = "language";
char const * const s_valLanguageShortcuts                                       = "languageShortcuts";
char const * const s_valLastConfigDir                                           = "lastConfigDir";
char const * const s_valLastOpenDir                                             = "lastOpenDir";
char const * const s_valLastOutputDir                                           = "lastOutputDir";
char const * const s_valLastUpdateCheck                                         = "lastUpdateCheck";
char const * const s_valMaximumConcurrentJobs                                   = "maximumConcurrentJobs";
char const * const s_valMediaInfoExe                                            = "mediaInfoExe";
char const * const s_valMergeAddBlurayCovers                                    = "mergeAddBlurayCovers";
char const * const s_valMergeAddingAppendingFilesPolicy                         = "mergeAddingAppendingFilesPolicy";
char const * const s_valMergeAlwaysAddDroppedFiles                              = "mergeAlwaysAddDroppedFiles";
char const * const s_valMergeAlwaysCreateNewSettingsForVideoFiles               = "mergeAlwaysCreateNewSettingsForVideoFiles";
char const * const s_valMergeAlwaysShowOutputFileControls                       = "mergeAlwaysShowOutputFileControls";
char const * const s_valMergeAttachmentsAlwaysSkipForExistingName               = "mergeAttachmentsAlwaysSkipForExistingName";
char const * const s_valMergeDefaultCommandLineEscapeMode                       = "mergeDefaultCommandLineEscapeMode";
char const * const s_valMergeDragAndDropDirectoriesPolicy                       = "mergeDragAndDropDirectoriesPolicy";
char const * const s_valMergeDragAndDropFilesPolicy                             = "mergeDragAndDropFilesPolicy";
char const * const s_valMergeEnableDialogNormGainRemoval                        = "mergeEnableDialogNormGainRemoval";
char const * const s_valMergeEnsureAtLeastOneTrackEnabled                       = "mergeEnsureAtLeastOneTrackEnabled";
char const * const s_valMergeFileColors                                         = "mergeFileColors";
char const * const s_valMergeLastAddingAppendingDecision                        = "mergeLastAddingAppendingDecision";
char const * const s_valMergeLastDragAndDropFilesDecision                       = "mergeLastDragAndDropFilesDecision";
char const * const s_valMergeLastFixedOutputDirs                                = "mergeLastFixedOutputDirs";
char const * const s_valMergeLastOutputDirs                                     = "mergeLastOutputDirs";
char const * const s_valMergeLastRelativeOutputDirs                             = "mergeLastRelativeOutputDirs";
char const * const s_valMergePredefinedAudioTrackNames                          = "mergePredefinedAudioTrackNames";
char const * const s_valMergePredefinedSplitDurations                           = "mergePredefinedSplitDurations";
char const * const s_valMergePredefinedSplitSizes                               = "mergePredefinedSplitSizes";
char const * const s_valMergePredefinedSubtitleTrackNames                       = "mergePredefinedSubttitleTrackNames";
char const * const s_valMergePredefinedTrackNames                               = "mergePredefinedTrackNames";
char const * const s_valMergePredefinedVideoTrackNames                          = "mergePredefinedVideoTrackNames";
char const * const s_valMergeReconstructSequencesWhenAdding                     = "mergeReconstructSequencesWhenAdding";
char const * const s_valMergeSetDestinationFromDirectory                        = "mergeSetDestinationFromDirectory";
char const * const s_valMergeSetDestinationFromTitle                            = "mergeSetDestinationFromTitle";
char const * const s_valMergeShowDNDZones                                       = "mergeShowDNDZones";
char const * const s_valMergeSortFilesTracksByTypeWhenAdding                    = "mergeSortFilesTracksByTypeWhenAdding";
char const * const s_valMergeTrackPropertiesLayout                              = "mergeTrackPropertiesLayout";
char const * const s_valMergeUseFileAndTrackColors                              = "mergeUseFileAndTrackColors";
char const * const s_valMergeUseVerticalInputLayout                             = "mergeUseVerticalInputLayout";
char const * const s_valMergeWarnMissingAudioTrack                              = "mergeWarnMissingAudioTrack";
char const * const s_valMinimumPlaylistDuration                                 = "minimumPlaylistDuration";
char const * const s_valMode                                                    = "mode";
char const * const s_valName                                                    = "name";
char const * const s_valNumRecentlyUsedStringsToRemember                        = "numRecentlyUsedStringsToRemember";
char const * const s_valOftenUsedCharacterSets                                  = "oftenUsedCharacterSets";
char const * const s_valOftenUsedCharacterSetsOnly                              = "oftenUsedCharacterSetsOnly";
char const * const s_valOftenUsedCountries                                      = "oftenUsedCountries";     // replaced by oftenUsedRegions in v51
char const * const s_valOftenUsedCountriesOnly                                  = "oftenUsedCountriesOnly"; // replaced by oftenUsedRegionsOnly in v51
char const * const s_valOftenUsedLanguages                                      = "oftenUsedLanguages";
char const * const s_valOftenUsedLanguagesOnly                                  = "oftenUsedLanguagesOnly";
char const * const s_valOftenUsedRegions                                        = "oftenUsedRegions";
char const * const s_valOftenUsedRegionsOnly                                    = "oftenUsedRegionsOnly";
char const * const s_valOutputFileNamePolicy                                    = "outputFileNamePolicy";
char const * const s_valPriority                                                = "priority";
char const * const s_valProbeRangePercentage                                    = "probeRangePercentage";
char const * const s_valRecognizedTrackLanguagesInFileNames                     = "recognizedTrackLanguagesInFileNames";
char const * const s_valRegexForRecognizingForcedSubtitleNames                  = "regexForRecognizingForcedSubtitleNames";
char const * const s_valRelativeOutputDir                                       = "relativeOutputDir";
char const * const s_valRemoveOldJobs                                           = "removeOldJobs";
char const * const s_valRemoveOldJobsDays                                       = "removeOldJobsDays";
char const * const s_valRemoveOutputFileOnJobFailure                            = "removeOutputFileOnJobFailure";
char const * const s_valResetJobWarningErrorCountersOnExit                      = "resetJobWarningErrorCountersOnExit";
char const * const s_valScanForPlaylistsPolicy                                  = "scanForPlaylistsPolicy";
char const * const s_valSetAudioDelayFromFileName                               = "setAudioDelayFromFileName";
char const * const s_valShowDebuggingMenu                                       = "showDebuggingMenu";
char const * const s_valShowMoveUpDownButtons                                   = "showMoveUpDownButtons";
char const * const s_valShowOutputOfAllJobs                                     = "showOutputOfAllJobs";
char const * const s_valShowToolSelector                                        = "showToolSelector";
char const * const s_valSubtitlePolicy                                          = "subtitlePolicy";
char const * const s_valSwitchToJobOutputAfterStarting                          = "switchToJobOutputAfterStarting";
char const * const s_valTabPosition                                             = "tabPosition";
char const * const s_valTrackInfo                                               = "trackInfo";
char const * const s_valTrackName                                               = "trackName";
char const * const s_valType                                                    = "type";
char const * const s_valUiDisableHighDPIScaling                                 = "uiDisableHighDPIScaling";
char const * const s_valUiDisableToolTips                                       = "uiDisableToolTips";
char const * const s_valUiFontFamily                                            = "uiFontFamily";
char const * const s_valUiFontPointSize                                         = "uiFontPointSize";
char const * const s_valUiLocale                                                = "uiLocale";
char const * const s_valUiPalette                                               = "uiPalette";
char const * const s_valUiStayOnTop                                             = "uiStayOnTop";
char const * const s_valUniqueOutputFileNames                                   = "uniqueOutputFileNames";
char const * const s_valUseDefaultJobDescription                                = "useDefaultJobDescription";
char const * const s_valUseISO639_3Languages                                    = "useISO639_3Languages";
char const * const s_valUseLegacyFontMIMETypes                                  = "useLegacyFontMIMETypes";
char const * const s_valVerbosity                                               = "verbosity";
char const * const s_valVideoPolicy                                             = "videoPolicy";
char const * const s_valVolume                                                  = "volume";
char const * const s_valWarnBeforeAbortingJobs                                  = "warnBeforeAbortingJobs";
char const * const s_valWarnBeforeClosingModifiedTabs                           = "warnBeforeClosingModifiedTabs";
char const * const s_valWarnBeforeOverwriting                                   = "warnBeforeOverwriting";
char const * const s_valWhenToSetDefaultLanguage                                = "whenToSetDefaultLanguage";

} // anonymous namespace
