Source: mkvtoolnix
Section: video
Priority: optional
Maintainer: Moritz Bunkus <mo@bunkus.online>
Homepage: https://mkvtoolnix.download/
Standards-Version: 3.9.5
Build-Depends:
 debhelper (>= 9),
 rake (>= 10),
 zlib1g-dev, libflac-dev, libogg-dev, libvorbis-dev, libgmp-dev, libcmark-dev,
 libboost-dev (>= 1.74), libboost-filesystem-dev (>= 1.74),
 libboost-system-dev (>= 1.74), libgtest-dev, libfmt-dev (>= 8),
 libdvdread-dev, nlohmann-json3-dev, pkg-config, po4a, docbook-xsl, xsltproc,
 qt6-base-dev, qt6-base-dev-tools, qt6-multimedia-dev, libqt6svg6-dev

Package: mkvtoolnix
Architecture: any
Depends: ${shlibs:Depends}, ${misc:Depends}
Suggests: mkvtoolnix-gui
Description: Set of command-line tools to work with Matroska files
 Matroska is a new multimedia container format, based on EBML (Extensible
 Binary Meta Language), which is a kind of binary XML.
 These tools allow one to manipulate Matroska files.
  * mkvmerge is a tool to create Matroska files from other formats
  * mkvinfo allows one to get information about the tracks in a Matroska file
  * mkvextract can extract tracks from Matroska files to other formats
  * mkvpropedit can edit header and chapter information without remuxing
 .
 This package only provides command-line tools. For graphical tools, you
 should install the mkvtoolnix-gui package.

Package: mkvtoolnix-gui
Architecture: any
Depends: mkvtoolnix (= ${binary:Version}), ${shlibs:Depends}, ${misc:Depends},
 libqt6svg6, qt6-qpa-plugins | qt6-wayland
Suggests: mediainfo-gui
Replaces: mkvtoolnix (<= 4.6.0-1)
Description: Set of tools to work with Matroska files - GUI frontend
 Matroska is a new multimedia container format, based on EBML (Extensible
 Binary Meta Language), which is a kind of binary XML.
 .
 The mkvtoolnix-gui tool is a graphical user interface for the mkvmerge
 program, which allows one to create Matroska files from other formats.

Package: mkvtoolnix-dbg
Section: debug
Priority: extra
Architecture: any
Depends: mkvtoolnix (= ${binary:Version}), ${misc:Depends}
Description: debugging symbols for mkvtoolnix
 Matroska is a new multimedia container format, based on EBML (Extensible
 Binary Meta Language), which is a kind of binary XML.
 These tools allow one to manipulate Matroska files.
  * mkvmerge is a tool to create Matroska files from other formats
  * mkvinfo allows one to get information about the tracks in a Matroska file
  * mkvextract can extract tracks from Matroska files to other formats
  * mkvpropedit can edit header and chapter information without remuxing
 .
 This package contains the debugging symbols for the command-line tools
 of mkvtoolnix.

Package: mkvtoolnix-gui-dbg
Section: debug
Priority: extra
Architecture: any
Depends: mkvtoolnix-gui (= ${binary:Version}), ${misc:Depends}
Description: debugging symbols for mkvtoolnix-gui
 Matroska is a new multimedia container format, based on EBML (Extensible
 Binary Meta Language), which is a kind of binary XML.
 These tools allow one to manipulate Matroska files.
  * mkvmerge is a tool to create Matroska files from other formats
  * mkvinfo allows one to get information about the tracks in a Matroska file
  * mkvextract can extract tracks from Matroska files to other formats
  * mkvpropedit can edit header and chapter information without remuxing
 .
 This package contains the debugging symbols for the GUI of
 mkvtoolnix-gui.
