#!/usr/bin/ruby -w

# T_573teletext_subs_data_unit_id_0x02_and_multiple_magazines
describe "mkvmerge / DVB teletext subtitles with data unit ID 0x02 and various magazines"
test_merge "data/ts/teletext_subs_data_unit_id_0x02_and_multiple_magazines.ts"
